package org.zkoss.chart;

import java.util.List;
import java.util.Map;

import org.zkoss.chart.model.ChartsModel;
import org.zkoss.chart.options3D.Options3D;
import org.zkoss.chart.plotOptions.PlotOptions;
import org.zkoss.chart.plotOptions.SeriesPlotOptions;
import org.zkoss.zk.ui.UiException;
import org.zkoss.zul.ChartModel;

/**
 * Charts' API interface.
 *
 * @author rudyhuang
 * @since 3.0.2
 */
public interface ChartsApi {
	/**
	 * Returns the plot data which is a JSON object to collect all chart's
	 * options for Charts JS.
	 */
	PlotData getPlotData();

	/**
	 * Returns the drilldown object.
	 */
	Drilldown getDrilldown();

	/**
	 * Sets the default colors for the chart's series. When all colors are used,
	 * new colors are pulled from the start again.
	 */
	void setColors(List<Color> colors);

	/**
	 * Sets the default colors for the chart's series. When all colors are used,
	 * new colors are pulled from the start again.
	 */
	void setColors(String... source);

	/**
	 * Returns the default colors for the chart's series. When all colors are
	 * used, new colors are pulled from the start again.
	 */
	List<Color> getColors();

	/**
	 * Sets the theme options for all Charts in the same browser page.
	 */
	void setTheme(Theme theme);

	/**
	 * Returns the global theme for all Charts.
	 * <p>
	 * Default is null
	 */
	Theme getTheme();

	/**
	 * Returns the chart model associated with this chart, or null if this chart
	 * is not associated with any chart data model.
	 */
	ChartsModel getModel();

	/**
	 * Sets the chart model associated with this chart. If a non-null model is
	 * assigned, no matter whether it is the same as the previous, it will
	 * always cause re-render.
	 *
	 * @param model
	 *            the chart model to associate, or null to dissociate any
	 *            previous model.
	 * @exception UiException
	 *                if failed to initialize with the model
	 */
	void setModel(ChartsModel model);

	void setModel(ChartModel model);

	/**
	 * Set the chart's type (Charts.PIE, Charts.BAR, Charts.LINE, etc.).
	 * <p>
	 * Default: line.
	 */
	void setType(String type);

	/**
	 * Get the chart's type.
	 * <p>
	 * Default: line.
	 */
	String getType();

	/**
	 * Sets when using multiple axis, the ticks of two or more opposite axes
	 * will automatically be aligned by adding ticks to the axis or axes with
	 * the least ticks. This can be prevented by setting <code>alignTicks</code>
	 * to false. If the grid lines look messy, it's a good idea to hide them for
	 * the secondary axis by setting <code>gridLineWidth</code> to 0.
	 *
	 * @param alignTicks
	 *            when using multiple axis, the ticks of two or more opposite
	 *            axes will automatically be aligned by adding ticks to the axis
	 *            or axes with the least ticks
	 * @see Chart#setAlignTicks(boolean)
	 */
	void setAlignTicks(boolean alignTicks);

	/**
	 * Returns when using multiple axis, the ticks of two or more opposite axes
	 * will automatically be aligned by adding ticks to the axis or axes with
	 * the least ticks. This can be prevented by setting <code>alignTicks</code>
	 * to false. If the grid lines look messy, it's a good idea to hide them for
	 * the secondary axis by setting <code>gridLineWidth</code> to 0.
	 * <p>
	 * Default: true.
	 *
	 * @see Chart#isAlignTicks()
	 */
	boolean isAlignTicks();

	/**
	 * Animation can be disabled throughout the chart by setting it to false
	 * here. It can be overridden for each individual API method as a function
	 * parameter. The only animation not affected by this option is the initial
	 * series animation, see {@link SeriesPlotOptions#setAnimation(boolean)}.
	 *
	 * @see Chart#setAnimation(boolean)
	 * @see Animation
	 */
	void setAnimation(boolean animation);

	/**
	 * Returns whether enable the animation
	 * <p>
	 * Default: true
	 */
	boolean isAnimation();

	/**
	 * Sets to ignore the hidden series, if true, the axes will scale to the
	 * remaining visible series once one series is hidden. If false, hiding and
	 * showing a series will not affect the axes or the other series. For
	 * stacks, once one series within the stack is hidden, the rest of the stack
	 * will close in around it even if the axis is not affected.
	 *
	 * @param ignoreHiddenSeries
	 *            if true, the axes will scale to the remaining visible series
	 *            once one series is hidden
	 */
	void setIgnoreHiddenSeries(boolean ignoreHiddenSeries);

	/**
	 * Returns if true, the axes will scale to the remaining visible series once
	 * one series is hidden. If false, hiding and showing a series will not
	 * affect the axes or the other series. For stacks, once one series within
	 * the stack is hidden, the rest of the stack will close in around it even
	 * if the axis is not affected.
	 * <p>
	 * Default: true.
	 */
	boolean isIgnoreHiddenSeries();

	/**
	 * Sets the background color for the outer chart area.
	 *
	 * @param backgroundColor
	 *            the background color for the outer chart area
	 */
	void setBackgroundColor(String backgroundColor);

	/**
	 * Sets the background gradient for the outer chart area.
	 */
	void setBackgroundColor(LinearGradient gradient);

	/**
	 * Sets the background gradient for the outer chart area.
	 */
	void setBackgroundColor(RadialGradient gradient);

	/**
	 * Sets the background gradient for the outer chart area.
	 */
	void setBackgroundColor(Color color);

	/**
	 * Returns the background color for the outer chart area.
	 * <p>
	 * Default: "#FFFFFF".
	 *
	 * @see Chart#getBackgroundColor()
	 */
	Color getBackgroundColor();

	/**
	 * Sets the color of the outer chart border. The border is painted using
	 * vector graphic techniques to allow rounded corners.
	 */
	void setBorderColor(String borderColor);

	/**
	 * Sets the color of the outer chart border. The border is painted using
	 * vector graphic techniques to allow rounded corners.
	 */
	void setBorderColor(LinearGradient borderColor);

	/**
	 * Sets the color of the outer chart border. The border is painted using
	 * vector graphic techniques to allow rounded corners.
	 */
	void setBorderColor(RadialGradient borderColor);

	/**
	 * Sets the color of the outer chart border. The border is painted using
	 * vector graphic techniques to allow rounded corners.
	 */
	void setBorderColor(Color borderColor);

	/**
	 * Returns the color of the outer chart border. The border is painted using
	 * vector graphic techniques to allow rounded corners.
	 * <p>
	 * Default: "#4572A7".
	 *
	 * @see Chart#getBorderColor()
	 */
	Color getBorderColor();

	/**
	 * Sets the corner radius of the outer chart border. In export, the radius
	 * defaults to 0.
	 * <p>
	 * Default: 5.
	 *
	 * @param borderRadius
	 *            the corner radius of the outer chart border
	 * @see Chart#setBorderRadius(Number)
	 */
	void setBorderRadius(Number borderRadius);

	/**
	 * Returns the corner radius of the outer chart border. In export, the
	 * radius defaults to 0.
	 * <p>
	 * Default: 5.
	 *
	 * @see Chart#getBorderRadius()
	 */
	Number getBorderRadius();

	/**
	 * Sets the pixel width of the outer chart border. The border is painted
	 * using vector graphic techniques to allow rounded corners.
	 *
	 * @see Chart#setBorderWidth(Number)
	 */
	void setBorderWidth(Number borderWidth);

	/**
	 * Returns the pixel width of the outer chart border. The border is painted
	 * using vector graphic techniques to allow rounded corners.
	 * <p>
	 * Default: 0.
	 *
	 * @see Chart#getBorderWidth()
	 */
	Number getBorderWidth();

	/**
	 * Sets an explicit width for the chart. By default the width is calculated
	 * from the offset width of the containing element.
	 *
	 * @param width
	 *            an explicit width for the chart
	 * @see Chart#setWidth(Number)
	 */
	void setWidth(Number width);

	/**
	 * Returns an explicit width for the chart. By default the width is
	 * calculated from the offset width of the containing element.
	 * <p>
	 * Default: null.
	 *
	 * @see Chart#getWidth()
	 */
	Number getWidth();

	/**
	 * Sets an explicit height for the chart. By default the height is
	 * calculated from the offset height of the containing element, or 400
	 * pixels if the containing element's height is 0.
	 *
	 * @param height
	 *            an explicit height for the chart
	 * @see Chart#setHeight(Number)
	 */
	void setHeight(Number height);

	/**
	 * Sets an explicit height for the chart. If given a percentage string
	 * (for example '56%'), the height is given as the percentage of the actual
	 * chart width. This allows for preserving the aspect ratio across responsive sizes.
	 * @since 7.2.1.0
	 */
	void setHeight(String height);

	/**
	 * Returns an explicit height for the chart. By default the height is
	 * calculated from the offset height of the containing element, or 400
	 * pixels if the containing element's height is 0.
	 * <p>
	 * Default: null.
	 *
	 * @see Chart#getHeight()
	 */
	Number getHeight();

	/**
	 * Sets whether to invert the axes so that the x axis is vertical and y axis
	 * is horizontal. When true, the x axis is reversed by default. If a bar
	 * series is present in the chart, it will be inverted automatically.
	 *
	 * @see Chart#setInverted(boolean)
	 */
	void setInverted(boolean inverted);

	/**
	 * Returns whether to invert the axes so that the x axis is vertical and y
	 * axis is horizontal. When true, the x axis is reversed by default. If a
	 * bar series is present in the chart, it will be inverted automatically.
	 * <p>
	 * Default: false.
	 *
	 * @see Chart#isInverted()
	 */
	boolean isInverted();

	/**
	 * Sets the margin between the outer edge of the chart and the plot area.
	 * The numbers in the array designate top, right, bottom and left
	 * respectively. Use the methods of {@link #setMarginTop(Number)},
	 * {@link #setMarginRight(Number)}, {@link #setMarginBottom(Number)} and
	 * {@link #setMarginLeft(Number)} for shorthand setting of one option.
	 * <p>
	 * Default: null.
	 *
	 * @see #setMargin(Number[])
	 */
	void setMargin(String margins);

	/**
	 * Sets the margin between the outer edge of the chart and the plot area.
	 *
	 */
	void setMargin(Number margin);

	/**
	 * Sets the margin between the outer edge of the chart and the plot area.
	 * The numbers in the array designate top, right, bottom and left
	 * respectively. Use the methods of {@link #setMarginTop(Number)},
	 * {@link #setMarginRight(Number)}, {@link #setMarginBottom(Number)} and
	 * {@link #setMarginLeft(Number)} for shorthand setting of one option.
	 * <p>
	 * Default: null.
	 *
	 * @see #setMargin(String)
	 * @see Chart#setMargin(Number[])
	 */
	void setMargin(Number[] margin);

	/**
	 * Returns the margin between the outer edge of the chart and the plot area.
	 * The numbers in the array designate top, right, bottom and left
	 * respectively. Use the methods of {@link #setMarginTop(Number)},
	 * {@link #setMarginRight(Number)}, {@link #setMarginBottom(Number)} and
	 * {@link #setMarginLeft(Number)} for shorthand setting of one option.
	 * <p>
	 * Default: null.
	 *
	 * @see Chart#getMargin()
	 */
	Number[] getMargin();

	/**
	 * Sets the margin between the bottom outer edge of the chart and the plot
	 * area. Use this to set a fixed pixel value for the margin as opposed to
	 * the default dynamic margin. See also {@link #setSpacingBottom(Number)}.
	 *
	 * @param marginBottom
	 *            the margin between the bottom outer edge of the chart and the
	 *            plot area
	 * @see Chart#setMarginBottom(Number)
	 */
	void setMarginBottom(Number marginBottom);

	/**
	 * Returns the margin between the bottom outer edge of the chart and the
	 * plot area.
	 * <p>
	 * Default: null.
	 *
	 * @see Chart#getMarginBottom()
	 */
	Number getMarginBottom();

	/**
	 * Sets the margin between the left outer edge of the chart and the plot
	 * area. Use this to set a fixed pixel value for the margin as opposed to
	 * the default dynamic margin. See also {@link #setSpacingLeft(Number)}
	 *
	 * @param marginLeft
	 *            the margin between the left outer edge of the chart and the
	 *            plot area
	 * @see Chart#setMarginLeft(Number)
	 */
	void setMarginLeft(Number marginLeft);

	/**
	 * Returns the margin between the left outer edge of the chart and the plot
	 * area.
	 * <p>
	 * Default: null.
	 *
	 * @see Chart#getMarginLeft()
	 */
	Number getMarginLeft();

	/**
	 * Sets the margin between the right outer edge of the chart and the plot
	 * area. Use this to set a fixed pixel value for the margin as opposed to
	 * the default dynamic margin. See also {@link #setSpacingRight(Number)}.
	 *
	 * @param marginRight
	 *            the margin between the right outer edge of the chart and the
	 *            plot area
	 * @see Chart#setMarginRight(Number)
	 */
	void setMarginRight(Number marginRight);

	/**
	 * Returns the margin between the right outer edge of the chart and the plot
	 * area.
	 * <p>
	 * Default: null.
	 *
	 * @see Chart#getMarginRight()
	 */
	Number getMarginRight();

	/**
	 * Sets the margin between the top outer edge of the chart and the plot
	 * area. Use this to set a fixed pixel value for the margin as opposed to
	 * the default dynamic margin. See also {@link #setSpacingTop(Number)}.
	 *
	 * @param marginTop
	 *            the margin between the top outer edge of the chart and the
	 *            plot area
	 * @see Chart#setMarginTop(Number)
	 */
	void setMarginTop(Number marginTop);

	/**
	 * Returns the margin between the top outer edge of the chart and the plot
	 * area.
	 * <p>
	 * Default: null.
	 *
	 * @see Chart#getMarginTop()
	 */
	Number getMarginTop();

	/**
	 * Sets equivalent to {@link #setZoomType(String)}, but for multitouch
	 * gestures only. By default, the <code>pinchType</code> is the same as the
	 * <code>zoomType</code> setting. However, pinching can be enabled
	 * separately in some cases, for example in stock charts where a mouse drag
	 * pans the chart, while pinching is enabled.
	 * <p>
	 * Default: null.
	 *
	 * @see #setZoomType(String)
	 * @see Chart#setPinchType(String)
	 */
	void setPinchType(String pinchType);

	/**
	 * Returns equivalent to {@link #getZoomType()}, but for multitouch gestures
	 * only. By default, the <code>pinchType</code> is the same as the
	 * <code>zoomType</code> setting. However, pinching can be enabled
	 * separately in some cases, for example in stock charts where a mouse drag
	 * pans the chart, while pinching is enabled.
	 * <p>
	 * Default: null.
	 *
	 * @see Chart#getPinchType()
	 */
	String getPinchType();

	/**
	 * Sets the background color for the plot area.
	 */
	void setPlotBackgroundColor(String plotBackgroundColor);

	/**
	 * Sets the background color for the plot area.
	 */
	void setPlotBackgroundColor(LinearGradient plotBackgroundColor);

	/**
	 * Sets the background color for the plot area.
	 */
	void setPlotBackgroundColor(RadialGradient plotBackgroundColor);

	/**
	 * Sets the background color for the plot area.
	 */
	void setPlotBackgroundColor(Color plotBackgroundColor);

	/**
	 * Returns the background color for the plot area.
	 * <p>
	 * Default: null.
	 *
	 * @see Chart#getPlotBackgroundColor()
	 */
	Color getPlotBackgroundColor();

	/**
	 * Sets the URL for an image to use as the plot background. To set an image
	 * as the background for the entire chart, set a CSS background image to the
	 * container element. Note that for the image to be applied to exported
	 * charts, its URL needs to be accessible by the export server.
	 *
	 * @param plotBackgroundImage
	 *            the URL for an image to use as the plot background
	 * @see Chart#setPlotBackgroundImage(String)
	 */
	void setPlotBackgroundImage(String plotBackgroundImage);

	/**
	 * Returns the URL for an image to use as the plot background.
	 * <p>
	 * Default: null.
	 *
	 * @see Chart#getPlotBackgroundImage()
	 */
	String getPlotBackgroundImage();

	/**
	 * Sets the color of the inner chart or plot area border.
	 */
	void setPlotBorderColor(String plotBorderColor);

	/**
	 * Sets the color of the inner chart or plot area border.
	 */
	void setPlotBorderColor(LinearGradient plotBorderColor);

	/**
	 * Sets the color of the inner chart or plot area border.
	 */
	void setPlotBorderColor(RadialGradient plotBorderColor);

	/**
	 * Sets the color of the inner chart or plot area border.
	 */
	void setPlotBorderColor(Color plotBorderColor);

	/**
	 * Returns the color of the inner chart or plot area border.
	 * <p>
	 * Default: "#C0C0C0".
	 *
	 * @see Chart#getPlotBorderColor()
	 */
	Color getPlotBorderColor();

	/**
	 * Sets the pixel width of the plot area border.
	 *
	 * @param plotBorderWidth
	 *            the pixel width of the plot area border
	 * @see Chart#setPlotBorderWidth(Number)
	 */
	void setPlotBorderWidth(Number plotBorderWidth);

	/**
	 * Returns the pixel width of the plot area border.
	 * <p>
	 * Default: 0.
	 *
	 * @see Chart#getPlotBorderWidth()
	 */
	Number getPlotBorderWidth();

	/**
	 * Sets whether to apply a drop shadow to the plot area. Requires that
	 * {@link #setPlotBackgroundColor(String)} be set.
	 *
	 * @param plotShadow
	 *            whether to apply a drop shadow to the plot area
	 * @see Chart#setPlotShadow(boolean)
	 */
	void setPlotShadow(boolean plotShadow);

	/**
	 * Returns whether to apply a drop shadow to the plot area. Requires that
	 * {@link #setPlotBackgroundColor(String)} be set.
	 * <p>
	 * Default: false.
	 *
	 * @see Chart#isPlotShadow()
	 */
	boolean isPlotShadow();

	/**
	 * Sets to true to enable the cartesian charts like line, spline, area and
	 * column are transformed into the polar coordinate system.
	 *
	 * @param polar
	 *            when true, cartesian charts like line, spline, area and column
	 *            are transformed into the polar coordinate system
	 * @see Chart#setPolar(boolean)
	 */
	void setPolar(boolean polar);

	/**
	 * Returns whether transforms into the polar coordinate system
	 * <p>
	 * Default: false.
	 *
	 * @see Chart#isPolar()
	 */
	boolean isPolar();

	/**
	 * Sets whether to reflow the chart to fit the width of the container div on
	 * resizing the window.
	 *
	 * @param reflow
	 *            whether to reflow the chart to fit the width of the container
	 *            div on resizing the window
	 * @see Chart#setReflow(boolean)
	 */
	void setReflow(boolean reflow);

	/**
	 * Returns whether to reflow the chart to fit the width of the container div
	 * on resizing the window.
	 * <p>
	 * Default: true.
	 *
	 * @see Chart#isReflow()
	 */
	boolean isReflow();

	/**
	 * Sets the background color of the marker square when selecting (zooming in
	 * on) an area of the chart.
	 */
	void setSelectionMarkerFill(String selectionMarkerFill);

	/**
	 * Sets the background color of the marker square when selecting (zooming in
	 * on) an area of the chart.
	 */
	void setSelectionMarkerFill(LinearGradient selectionMarkerFill);

	/**
	 * Sets the background color of the marker square when selecting (zooming in
	 * on) an area of the chart.
	 */
	void setSelectionMarkerFill(RadialGradient selectionMarkerFill);

	/**
	 * Sets the background color of the marker square when selecting (zooming in
	 * on) an area of the chart.
	 */
	void setSelectionMarkerFill(Color selectionMarkerFill);

	/**
	 * Returns the background color of the marker square when selecting (zooming
	 * in on) an area of the chart.
	 * <p>
	 * Default: "rgba(69,114,167,0.25)".
	 *
	 * @see Chart#getSelectionMarkerFill()
	 */
	Color getSelectionMarkerFill();

	/**
	 * Sets whether to apply a drop shadow to the outer chart area. Requires
	 * that backgroundColor be set.
	 *
	 * @param shadow
	 *            whether to apply a drop shadow to the outer chart area
	 * @see Chart#setShadow(boolean)
	 */
	void setShadow(boolean shadow);

	/**
	 * Returns whether to apply a drop shadow to the outer chart area. Requires
	 * that backgroundColor be set.
	 * <p>
	 * Default: false.
	 *
	 * @see Chart#isShadow()
	 */
	boolean isShadow();

	/**
	 * Sets whether to show the axes initially. This only applies to empty
	 * charts where series are added dynamically, as axes are automatically
	 * added to cartesian series.
	 *
	 * @param showAxes
	 *            whether to show the axes initially
	 * @see Chart#setShowAxes(boolean)
	 */
	void setShowAxes(boolean showAxes);

	/**
	 * Returns whether to show the axes initially. This only applies to empty
	 * charts where series are added dynamically, as axes are automatically
	 * added to cartesian series.
	 * <p>
	 * Default: false.
	 *
	 * @see Chart#isShowAxes()
	 */
	boolean isShowAxes();

	/**
	 * Sets the distance between the outer edge of the chart and the content,
	 * like title, legend, axis title or labels. The numbers in the array
	 * designate top, right, bottom and left respectively. Use the methods of
	 * {@link #setSpacingTop(Number)}, {@link #setSpacingRight(Number)},
	 * {@link #setSpacingBottom(Number)} and {@link #setSpacingLeft(Number)} for
	 * shorthand setting of one option.
	 *
	 * @param spacings
	 *            the distance between the outer edge of the chart and the
	 *            content, like title, legend, axis title or labels
	 * @see #setSpacing(Number[])
	 */
	void setSpacing(String spacings);

	/**
	 * Sets the distance between the outer edge of the chart and the content,
	 * like title, legend, axis title or labels. The numbers in the array
	 * designate top, right, bottom and left respectively. Use the methods of
	 * {@link #setSpacingTop(Number)}, {@link #setSpacingRight(Number)},
	 * {@link #setSpacingBottom(Number)} and {@link #setSpacingLeft(Number)} for
	 * shorthand setting of one option.
	 *
	 * @param spacing
	 *            the distance between the outer edge of the chart and the
	 *            content, like title, legend, axis title or labels
	 * @see Chart#setSpacing(Number[])
	 */
	void setSpacing(Number[] spacing);

	/**
	 * Returns the distance between the outer edge of the chart and the content,
	 * like title, legend, axis title or labels. The numbers in the array
	 * designate top, right, bottom and left respectively.
	 * <p>
	 * Default: [10, 10, 15, 10].
	 */
	Number[] getSpacing();

	/**
	 * Sets the space between the bottom edge of the chart and the content (plot
	 * area, axis title and labels, title, subtitle or legend in top position).
	 * </p>
	 *
	 * @param spacingBottom
	 *            The space between the bottom edge of the chart and the content
	 *            (plot area, axis title and labels, title, subtitle or legend
	 *            in top position)
	 * @see Chart#setSpacingBottom(Number)
	 */
	void setSpacingBottom(Number spacingBottom);

	/**
	 * Returns the space between the bottom edge of the chart and the content
	 * (plot area, axis title and labels, title, subtitle or legend in top
	 * position).
	 * <p>
	 * Default: 15.
	 *
	 * @see Chart#getSpacingBottom()
	 */
	Number getSpacingBottom();

	/**
	 * Sets the space between the left edge of the chart and the content (plot
	 * area, axis title and labels, title, subtitle or legend in top position).
	 * </p>
	 *
	 * @param spacingLeft
	 *            The space between the left edge of the chart and the content
	 *            (plot area, axis title and labels, title, subtitle or legend
	 *            in top position)
	 * @see Chart#setSpacingLeft(Number)
	 */
	void setSpacingLeft(Number spacingLeft);

	/**
	 * Returns the space between the left edge of the chart and the content
	 * (plot area, axis title and labels, title, subtitle or legend in top
	 * position).
	 * <p>
	 * Default: 10.
	 *
	 * @see Chart#getSpacingLeft()
	 */
	Number getSpacingLeft();

	/**
	 * Sets the space between the right edge of the chart and the content (plot
	 * area, axis title and labels, title, subtitle or legend in top position).
	 *
	 * @param spacingRight
	 *            The space between the right edge of the chart and the content
	 *            (plot area, axis title and labels, title, subtitle or legend
	 *            in top position)
	 * @see Chart#setSpacingRight(Number)
	 */
	void setSpacingRight(Number spacingRight);

	/**
	 * Returns the space between the right edge of the chart and the content
	 * (plot area, axis title and labels, title, subtitle or legend in top
	 * position).
	 * <p>
	 * Default: 10.
	 *
	 * @see Chart#getSpacingRight()
	 */
	Number getSpacingRight();

	/**
	 * Sets the space between the top edge of the chart and the content (plot
	 * area, axis title and labels, title, subtitle or legend in top position).
	 *
	 * @param spacingTop
	 *            The space between the top edge of the chart and the content
	 *            (plot area, axis title and labels, title, subtitle or legend
	 *            in top position)
	 * @see Chart#setSpacingTop(Number)
	 */
	void setSpacingTop(Number spacingTop);

	/**
	 * Returns the space between the top edge of the chart and the content (plot
	 * area, axis title and labels, title, subtitle or legend in top position).
	 * <p>
	 * Default: 10.
	 *
	 * @see Chart#getSpacingTop()
	 */
	Number getSpacingTop();

	/**
	 * Sets additional CSS styles to apply inline to the container
	 * <code>div</code>. Note that since the default font styles are applied in
	 * the renderer, it is ignorant of the individual chart options and must be
	 * set globally.
	 * <p>
	 * This method allows <b>simple</b> settings like
	 * <pre>{@code
	 * label.setStyle("fontWeight: 'bold'; color: 'white'; textShadow: '0px 1px 2px black'");
	 * }</pre>
	 * <p>
	 * In the format of CSS key:value, and ";" separating multiple key:value pairs,
	 * if you have more a complex CSS styling, we recommend using {@link #setStyle(Map)}
	 * @param style one or many CSS key:value pairs in a single line string format
	 */
	void setStyle(String style);

	/**
	 * Sets additional CSS styles to apply inline to the container
	 * <code>div</code>. Note that since the default font styles are applied in
	 * the renderer, it is ignorant of the individual chart options and must be
	 * set globally.
	 * <p>
	 * This method allows <b>complex</b> styling like:
	 * <pre>
	 * {@code
	 * Map<String, Object> styles = new HashMap<String, Object>();
	 * styles.put("fill", "#E8E8EA");
	 * styles.put("stroke", "#6E6E78");
	 * styles.put("stroke-width", "1");
	 * styles.put("r", "0");
	 * Map<String, Object> states = new HashMap<String, Object>();
	 * Map<String, String> hover = new HashMap<String, String>();
	 * hover.put("fill", "#F4F4F6");
	 * hover.put("stroke", "#6E6E78");
	 * hover.put("stroke-width", "1");
	 * hover.put("r", "0");
	 * states.put("hover", hover);
	 * styles.put("states", states);
	 * Map<String, String> style = new HashMap<String, String>();
	 * style.put("color", "#4D5663");
	 * style.put("fontSize", "12px");
	 * style.put("fontFamily", "'Arial'");
	 * style.put("fontWeight", "normal");
	 * style.put("padding", "10px");
	 * styles.put("style", style);
	 * }</pre>
	 * which will generate the final results like:
	 * <pre>{@code
	 * {
	 *     "fill": "#E8E8EA",
	 *     "stroke": "#6E6E78",
	 *     "stroke-width": "1",
	 *     "r": "0",
	 *     "states":
	 *     {
	 *         "hover":
	 *         {
	 *             "fill": "#F4F4F6",
	 *             "stroke": "#6E6E78",
	 *             "stroke-width": "1",
	 *             "r": "0"
	 *         }
	 *     },
	 *     "style":
	 *     {
	 *         "color": "#4D5663",
	 *         "fontSize": "12px",
	 *         "fontFamily": "'Arial'",
	 *         "fontWeight": "normal",
	 *         "padding": "10px"
	 *     }
	 * }
	 * }</pre>
	 * @param <K> Type for the key, usually String
	 * @param <V> Type for the value, usually String. For nested maps, use Object or any other type.
	 * @param style the style object
	 */
	<K, V> void setStyle(Map<K, V> style);

	/**
	 * Returns additional CSS styles to apply inline to the container
	 * <code>div</code>. Note that since the default font styles are applied in
	 * the renderer, it is ignorant of the individual chart options and must be
	 * set globally. Defaults to:
	 *
	 * <pre>
	 * fontFamily: "Lucida Grande", "Lucida Sans Unicode", Verdana, Arial, Helvetica, sans-serif; fontSize: 12px
	 * </pre>
	 * <p>
	 * @param <K> type for the key
	 * <p>
	 * if style was set using {@link #setStyle(String)}, type will be String
	 * <p>
	 * if style was set using {@link #setStyle(Map)}, type will be the same as the type from the key of the map
	 * <p>
	 * if style was not set, default style will be applied, and type will be String
	 * @param <V> type for the value
	 * <p>
	 * if style was set using {@link #setStyle(String)}, type will be String
	 * <p>
	 * if style was set using {@link #setStyle(Map)}, type will be the same as the type from the key of the map
	 * <p>
	 * if style was not set, default style will be applied, and type will be String
	 */
	<K, V> Map<K, V> getStyle();

	/**
	 * Sets decides in what dimentions the user can zoom by dragging the mouse.
	 * Can be one of <code>x</code>, <code>y</code> or <code>xy</code>.
	 *
	 * @param zoomType
	 *            decides in what dimentions the user can zoom by dragging the
	 *            mouse
	 * @see Chart#setZoomType(String)
	 */
	void setZoomType(String zoomType);

	/**
	 * Returns decides in what dimentions the user can zoom by dragging the
	 * mouse. Can be one of <code>x</code>, <code>y</code> or <code>xy</code>.
	 * <p>
	 * Default: null.
	 *
	 * @see Chart#getZoomType()
	 */
	String getZoomType();

	/**
	 * Returns the Tooltip options, if any. Otherwise, create a new one.
	 *
	 * @see Tooltip
	 */
	Tooltip getTooltip();

	/**
	 * Sets the Tooltip options
	 *
	 * @param tooltip
	 */
	void setTooltip(Tooltip tooltip);

	/**
	 * Returns the Chart options, if any. Otherwise, create a new one.
	 *
	 * @see Chart
	 */
	Chart getChart();

	/**
	 * Sets the no-data options
	 * @since 1.1.0
	 */
	void setNoData(NoData noData);

	/**
	 * Returns the no-data options
	 * @since 1.1.0
	 */
	NoData getNoData();

	/**
	 * Returns the Exporting options, if any. Otherwise, create a new one.
	 *
	 * @see Exporting
	 */
	Exporting getExporting();

	/**
	 * Sets the Exporting options.
	 *
	 * @see Exporting
	 */
	void setExporting(Exporting exporting);

	/**
	 * Sets the Credits options
	 *
	 * @param credits
	 */
	void setCredits(Credits credits);

	/**
	 * Returns the Credits options, if any. Otherwise, create a new one.
	 *
	 * @see Credits
	 */
	Credits getCredits();

	/**
	 * Sets the Chart options
	 *
	 * @param chart
	 */
	void setChart(Chart chart);

	/**
	 * Returns the pane at the index 0
	 */
	Pane getPane();

	/**
	 * Returns the pane from the given index
	 */
	Pane getPane(int index);

	/**
	 * Returns the size of the pane list
	 */
	int getPaneSize();

	/**
	 * Returns the Legend options, if any. Otherwise, create a new one.
	 *
	 * @see Legend
	 */
	Legend getLegend();

	/**
	 * Returns the Loading options, if any. Otherwise, create a new one.
	 *
	 * @see Loading
	 */
	Loading getLoading();

	/**
	 * Sets the Legend options
	 *
	 * @param legend
	 */
	void setLegend(Legend legend);

	/**
	 * Returns the labels options
	 */
	Labels getLabels();

	/**
	 * Sets the labels options
	 */
	void setLabels(Labels labels);

	/**
	 * Sets the navigation options
	 */
	void setNavigation(Navigation navigation);

	/**
	 * Returns the navigation options
	 */
	Navigation getNavigation();

	/**
	 * Returns the plotOptions option, if any. Otherwise, create a new one.
	 *
	 * @see PlotOptions
	 */
	PlotOptions getPlotOptions();

	/**
	 * Sets the plotOptions options
	 *
	 * @param plotOptions
	 */
	void setPlotOptions(PlotOptions plotOptions);

	/**
	 * Returns the Series options, if any. Otherwise, create a new one.
	 * <p>
	 * Returns the first series (index 0) by default.
	 *
	 * @see Series
	 * @see #getSeries(int)
	 */
	Series getSeries();

	/**
	 * Add a series to the chart after render time. Note that this method should
	 * never be used when adding data synchronously at chart render time, as it
	 * adds expense to the calculations and rendering. When adding data at the
	 * same time as the chart is initiated, add the series as a configuration
	 * option instead.
	 *
	 * @param series
	 */
	void addSeries(Series series);

	/**
	 * Add an axis to the chart after render time. Note that this method should
	 * never be used when adding data synchronously at chart render time, as it
	 * adds expense to the calculations and rendering. When adding data at the
	 * same time as the chart is initiated, add the axis as a configuration
	 * option instead.
	 *
	 * @param axis
	 */
	void addAxis(Axis axis);

	/**
	 * Returns the size of series
	 */
	int getSeriesSize();

	/**
	 * Returns the Series options from the given index, if any. Otherwise,
	 * create a new one.
	 *
	 * @see Series
	 */
	Series getSeries(int index);

	/**
	 * Returns the XAxis options, if any. Otherwise, create a new one.
	 * <p>
	 * Returns the first xAxis (index 0) by default.
	 *
	 * @see XAxis
	 * @see #getXAxis(int)
	 */
	XAxis getXAxis();

	/**
	 * Returns the XAxis options from the given index, if any. Otherwise, create
	 * a new one.
	 *
	 * @see XAxis
	 */
	XAxis getXAxis(int index);

	/**
	 * Returns the size of xAxis
	 */
	int getXAxisSize();

	/**
	 * Sets the xAxis with out putting it in an Array
	 */
	void setXAxis(XAxis xaxis);

	/**
	 * Returns the YAxis options, if any. Otherwise, create a new one.
	 * <p>
	 * Returns the first yAxis (index 0) by default.
	 *
	 * @see YAxis
	 * @see #getYAxis(int)
	 */
	YAxis getYAxis();

	/**
	 * Returns the YAxis options from the given index, if any. Otherwise, create
	 * a new one.
	 *
	 * @see YAxis
	 */
	YAxis getYAxis(int index);

	/**
	 * Returns the size of yAxis
	 */
	int getyAxisSize();

	/**
	 * Sets the yAxis with out putting it in an Array
	 */
	void setYAxis(YAxis yaxis);

	/**
	 * Returns the ZAxis options, if any. Otherwise, create a new one.
	 * <p>
	 * Returns the first zAxis (index 0) by default.
	 *
	 * @see ZAxis
	 * @see #getZAxis(int)
	 */
	ZAxis getZAxis();

	/**
	 * Returns the ZAxis options from the given index, if any. Otherwise, create
	 * a new one.
	 *
	 * @see ZAxis
	 */
	ZAxis getZAxis(int index);

	/**
	 * Returns the size of zAxis
	 */
	int getzAxisSize();

	/**
	 * Sets the zAxis with out putting it in an Array
	 */
	void setZAxis(ZAxis zaxis);

	/**
	 * Returns the options to render charts in 3 dimensions.
	 * <p>
	 * Default: null.
	 */
	Options3D getOptions3D();

	/**
	 * Sets the options to render charts in 3 dimensions.
	 *
	 * @param options3D
	 *            the options to render charts in 3 dimensions.
	 */
	void setOptions3D(Options3D options3D);

	/**
	 * Returns the ColorAxis options, if any. Otherwise, create a new one.
	 * <p>
	 * Returns the first colorAxis (index 0) by default.
	 *
	 * @see ColorAxis
	 * @see #getColorAxis(int)
	 */
	ColorAxis getColorAxis();

	/**
	 * Returns the ColorAxis options from the given index, if any. Otherwise, create a new one.
	 * <p>
	 *
	 * @see ColorAxis
	 * @see #getColorAxis()
	 * @since 7.2.1.0
	 */
	ColorAxis getColorAxis(int index);

	/**
	 * Sets the chart title.
	 */
	void setTitle(String title);

	/**
	 * Sets the Title options
	 *
	 * @param title
	 */
	void setTitle(Title title);

	/**
	 * Returns the Title options, if any. Otherwise, create a new one.
	 *
	 * @see Title
	 */
	Title getTitle();

	/**
	 * Returns the Subtitle options, if any. Otherwise, create a new one.
	 *
	 * @see Subtitle
	 */
	Subtitle getSubtitle();

	/**
	 * Sets the chart subtitle
	 *
	 * @param subtitle
	 */
	void setSubtitle(String subtitle);

	/**
	 * Sets the Subtitle options
	 *
	 * @param subtitle
	 */
	void setSubtitle(Subtitle subtitle);

	/**
	 * Sets whether selection event will zoom the chart to the selected area.
	 *
	 * @see Chart#setEnableZoomSelection(boolean)
	 */
	void setEnableZoomSelection(boolean enableZoomSelection);

	/**
	 * Returns whether selection event will zoom the chart to the selected area.
	 * <p>
	 * Default: true.
	 *
	 * @see Chart#isEnableZoomSelection()
	 */
	boolean isEnableZoomSelection();

	/**
	 * Returns the Annotation options, if any. Otherwise, create a new one.
	 * <p>
	 * Returns the first annotation (index 0) by default.
	 *
	 * @see Annotation
	 * @see #getAnnotation(int)
	 * @since 7.2.1.0
	 */
	Annotation getAnnotation();

	/**
	 * Returns the Annotation options from the given index, if any.
	 * Otherwise, create a new one.
	 * <p>
	 *
	 * @param index
	 * @see Annotation
	 * @since 7.2.1.0
	 */
	Annotation getAnnotation(int index);


	/**
	 * Returns the size of annotations
	 * @since 7.2.1.0
	 */
	int getAnnotationSize();

	/**
	 * Add an annotation to the chart after render time. Note that this method should
	 * never be used when adding data synchronously at chart render time, as it
	 * adds expense to the calculations and rendering. When adding data at the
	 * same time as the chart is initiated, add the annotation as a configuration
	 * option instead.
	 *
	 * @param annotation
	 * @since 7.2.1.0
	 */
	void addAnnotation(Annotation annotation);

	/**
	 * Sets the chart caption.
	 * @since 7.2.1.0
	 */
	void setCaption(String caption);

	/**
	 * Sets the Caption option.
	 * @since 7.2.1.0
	 */
	void setCaption(Caption caption);

	/**
	 * Returns the Caption options, if any. Otherwise, create a new one.
	 * @see Caption
	 * @since 7.2.1.0
	 */
	Caption getCaption();

	/**
	 * Returns a set of rules to apply for different screen or chart sizes.
	 * Each rule specifies additional chart options.
	 * @since 7.2.1.0
	 */
	Responsive getResponsive();

	/**
	 * Sets a set of rules to apply for different screen or chart sizes.
	 * Each rule specifies additional chart options.
	 * @since 7.2.1.0
	 */
	void setResponsive(Responsive responsive);
}
