/* Responsive.java

	Purpose:
		
	Description:
		
	History:
		Wed Dec 04 12:27:01 CST 2019, Created by rudyhuang

Copyright (C) 2019 Potix Corporation. All Rights Reserved.
*/
package org.zkoss.chart;

import java.util.LinkedList;
import java.util.List;

import org.zkoss.json.JavaScriptValue;

/**
 * Allows setting a set of rules to apply for different screen or chart sizes.
 * Each rule specifies additional chart options.
 * The rules are executed from the top down.
 *
 * @author rudyhuang
 * @since 7.2.1.0
 */
public class Responsive extends Optionable {
	private enum Attrs implements PlotAttribute {
		rules
	}

	/**
	 * Returns a rule for responsive settings at the index 0, if any.
	 * Otherwise, create a new one.
	 */
	public Rule getRule() {
		return getRule(0);
	}

	/**
	 * Returns a rule for responsive settings from the given index, if any.
	 * Otherwise, create a new one.
	 */
	public Rule getRule(int index) {
		List<Rule> list = (LinkedList<Rule>) getAttr(Attrs.rules);
		if (list == null) {
			list = new LinkedList<Rule>();
			setAttr(Attrs.rules, list);
		}

		int size = list.size();
		if (size <= index) {
			for (int i = index - size + 1; i > 0; i--) {
				list.add(new Rule());
			}
		}
		return list.get(index);
	}

	/**
	 * Adds the rule at the end of the rules list.
	 */
	public void addRule(Rule rule) {
		List<Rule> list = (LinkedList<Rule>) getAttr(Attrs.rules);
		if (list == null) {
			list = new LinkedList<Rule>();
			setAttr(Attrs.rules, list);
		}
		list.add(rule);
	}

	/**
	 * Returns the size of the rules list.
	 */
	public int getRuleSize() {
		List<Rule> list = (LinkedList<Rule>) getAttr(Attrs.rules);
		if (list == null)
			return 0;
		else
			return list.size();
	}

	/**
	 * A rule for responsive settings.
	 * @author rudyhuang
	 * @since 7.2.1.0
	 */
	public static class Rule extends Optionable {
		private enum Attrs implements PlotAttribute {
			chartOptions,
			condition
		}

		/**
		 * Returns a full set of chart options to apply as overrides to the general chart options, if any.
		 * Otherwise, create a new one.
		 * The chart options are applied when the given rule is active.
		 */
		public PlotData getChartOptions() {
			PlotData chartOptions = (PlotData) getAttr(Attrs.chartOptions);
			if (chartOptions == null) {
				chartOptions = new PlotData(null);
				setChartOptions(chartOptions);
			}
			return chartOptions;
		}

		/**
		 * Sets a full set of chart options to apply as overrides to the general chart options.
		 * The chart options are applied when the given rule is active.
		 *
		 * A special case is configuration objects that take arrays, for example xAxis,
		 * yAxis or series. For these collections, an id option is used to map the
		 * new option set to an existing object. If an existing object of the same
		 * id is not found, the item of the same index updated. So for example,
		 * setting chartOptions with two series items without an id, will cause
		 * the existing chart's two series to be updated with respective options.
		 */
		public void setChartOptions(PlotData chartOptions) {
			setAttr(Attrs.chartOptions, chartOptions);
		}

		/**
		 * Returns the condition the rule applies.
		 */
		public Condition getCondition() {
			Condition condition = (Condition) getAttr(Attrs.condition);
			if (condition == null) {
				condition = new Condition();
				setCondition(condition);
			}
			return condition;
		}

		/**
		 * Sets the condition the rule applies.
		 */
		public void setCondition(Condition condition) {
			setAttr(Attrs.condition, condition);
		}
	}

	/**
	 * Under which conditions the rule applies.
	 * @author rudyhuang
	 * @since 7.2.1.0
	 */
	public static class Condition extends Optionable {
		private enum Attrs implements PlotAttribute {
			callback,
			maxHeight,
			maxWidth,
			minHeight,
			minWidth
		}

		/**
		 * Returns a callback function to gain complete control on when the
		 * responsive rule applies. Return true if it applies.
		 * This opens for checking against other metrics than the chart size,
		 * for example the document size or other elements.
		 *
		 * Context: Highcharts.Chart
		 * <p>
		 * Default: null
		 */
		public JavaScriptValue getCallback() {
			return (JavaScriptValue) getAttr(Attrs.callback);
		}

		/**
		 * Sets a callback function to gain complete control on when the
		 * responsive rule applies. Return true if it applies.
		 * This opens for checking against other metrics than the chart size,
		 * for example the document size or other elements.
		 *
		 * Context: Highcharts.Chart
		 */
		public void setCallback(JavaScriptValue callback) {
			setAttr(Attrs.callback, callback);
		}

		/**
		 * Returns the value the responsive rule applies if the chart height is less than this.
		 * <p>
		 * Default: null
		 */
		public Number getMaxHeight() {
			return getAttr(Attrs.maxHeight, null).asNumber();
		}

		/**
		 * Sets the value the responsive rule applies if the chart height is less than this.
		 */
		public void setMaxHeight(Number maxHeight) {
			setAttr(Attrs.maxHeight, maxHeight);
		}

		/**
		 * Returns the value the responsive rule applies if the chart width is less than this.
		 * <p>
		 * Default: null
		 */
		public Number getMaxWidth() {
			return getAttr(Attrs.maxWidth, null).asNumber();
		}

		/**
		 * Sets the value the responsive rule applies if the chart width is less than this.
		 */
		public void setMaxWidth(Number maxWidth) {
			setAttr(Attrs.maxWidth, maxWidth);
		}

		/**
		 * Returns the value the responsive rule applies if the chart height is greater than this.
		 * <p>
		 * Default: 0
		 */
		public Number getMinHeight() {
			return getAttr(Attrs.minHeight, 0).asNumber();
		}

		/**
		 * Sets the value the responsive rule applies if the chart height is greater than this.
		 */
		public void setMinHeight(Number minHeight) {
			setAttr(Attrs.minHeight, minHeight);
		}

		/**
		 * Returns the value the responsive rule applies if the chart width is greater than this.
		 * <p>
		 * Default: 0
		 */
		public Number getMinWidth() {
			return getAttr(Attrs.minWidth, 0).asNumber();
		}

		/**
		 * Sets the value the responsive rule applies if the chart width is greater than this.
		 */
		public void setMinWidth(Number minWidth) {
			setAttr(Attrs.minWidth, minWidth);
		}
	}
}
