/** Series.java.

	Purpose:
		
	Description:
		
	History:
		2:06:58 PM Jan 9, 2014, Created by jumperchen

Copyright (C) 2014 Potix Corporation. All Rights Reserved.
 */
package org.zkoss.chart;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.zkoss.chart.OptionDataEvent.EventType;
import org.zkoss.chart.plotOptions.DataLabels;
import org.zkoss.chart.plotOptions.DragDrop;
import org.zkoss.chart.plotOptions.NetworkGraphLink;
import org.zkoss.chart.plotOptions.SeriesPlotOptions;
import org.zkoss.chart.plotOptions.TargetOptions;
import org.zkoss.chart.util.AnyVal;
import org.zkoss.chart.util.DeferredCall;
import org.zkoss.chart.util.DynamicalAttribute;
import org.zkoss.chart.util.JSFunction;
import org.zkoss.json.JSONObject;
import org.zkoss.lang.Generics;
import org.zkoss.lang.Objects;

/**
 * The Series object is the JavaScript representation of each line, area series,
 * pie etc. The object can be accessed in a number of ways. All series and point
 * event handlers give a reference to the series object. The chart object has a
 * series property that is a collection of all the chart's series. The point
 * objects also have the same reference. Another way to reference the series
 * programmatically is by id.
 * <p>
 * All the options in this class support {@link DynamicalAttribute}.
 * 
 * @author jumperchen
 */
public class Series extends Optionable implements OptionDataListener {
	private enum DAttrs implements PlotAttribute, DynamicalAttribute {
		animation,
		baseSeries,
		borderColor,
		borderRadius,
		borderWidth,
		center,
		color,
		colorAxis,
		colorKey,
		colorByPoint,
		colors,
		cursor,
		dataLabels,
		dragDrop,
		enableMouseTracking,
		endAngle,
		fillColor,
		gapSize,
		id,
		index,
		intervals,
		innerSize,
		legendIndex,
		layout,
		lineWidth,
		link,
		marker,
		maxSize,
		minSize,
		name,
		nodes,
		onSeries,
		overshoot,
		pointInterval,
		pointIntervalUnit,
		pointPadding,
		pointPlacement,
		pointsInInterval,
		pointStart,
		pointRange,
		rows,
		showInLegend,
		size,
		sizeBy,
		stack,
		startAngle,
		targetOptions,
		tooltip,
		type,
		xAxis,
		yAxis,
		zIndex,
		zoneAxis,
		zones
	}

	private enum Attrs implements PlotAttribute {
		data,
		visible,

		// internal use
		point
	}

	private SeriesPlotOptions plotOptions;

	public Series() {
	}

	/**
	 * Constructs the series with id and a list of point data.
	 */
	public Series(String id, List<Point> data) {
		setId(id);
		setData(data.toArray(new Point[0]));
	}

	/**
	 * Constructs the series with id and point data.
	 */
	public Series(String id, Point... data) {
		setId(id);
		setData(data);
	}

	/**
	 * Returns an id for the series.
	 * <p>
	 * Default: null.
	 */
	public String getId() {
		return getAttr(DAttrs.id, null).asString();
	}

	/**
	 * Sets an id for the series.
	 */
	public void setId(String id) {
		setAttr(DAttrs.id, id);
	}
	
	/**
	 * Returns whether to enable or disable the initial animation when a series is displayed.
	 * <p>
	 * Default: true.
	 * @since 7.2.1.0
	 */
	public boolean isAnimation() {
		return getAnimation() != Animation.NONE;
	}
	
	/**
	 * Sets whether to enable or disable the initial animation when a series is displayed.
	 * @see #setAnimation(Animation)
	 */
	public void setAnimation(boolean animation) {
		setAnimation(animation ? new Animation() : Animation.NONE);
	}

	/**
	 * Returns whether to enable or disable the initial animation when a series is displayed.
	 * <p>
	 * Default: an instance of {@link Animation}.
	 * @since 7.2.1.0
	 */
	public Animation getAnimation() {
		return (Animation) getAttr(DAttrs.animation, Animation.class).asValue();
	}

	/**
	 * Sets the initial animation when a series is displayed.
	 * Please note that this option only applies to the initial animation of the series itself.
	 * For other animations, see chart.animation and the animation parameter under the API methods.
	 * @see #setAnimation(boolean)
	 * @since 7.2.1.0
	 */
	public void setAnimation(Animation animation) {
		setAttr(DAttrs.animation, animation == null ? Animation.NONE : animation);
	}

	/**
	 * Returns an integer identifying the index to use for the base series,
	 * or a string representing the id of the series.
	 * <p>
	 * Default: null
	 * @since 7.2.1.0
	 */
	public Object getBaseSeries() {
		return getAttr(DAttrs.baseSeries);
	}

	/**
	 * Sets an integer identifying the index to use for the base series.
	 * @since 7.2.1.0
	 */
	public void setBaseSeries(int baseSeries) {
		setAttr(DAttrs.baseSeries, baseSeries);
	}

	/**
	 * Sets a string representing the id of the series to use for the base series.
	 * @since 7.2.1.0
	 */
	public void setBaseSeries(String baseSeries) {
		setAttr(DAttrs.baseSeries, baseSeries);
	}

	/**
	 * Returns the color of the border surrounding each slice.
	 * <p>
	 * Default: #ffffff
	 * @since 7.2.1.0
	 */
	public Color getBorderColor() {
		return (Color) getAttr(DAttrs.borderColor, new Color("#ffffff")).asValue();
	}

	/**
	 * Sets the color of the border surrounding each slice.
	 * @since 7.2.1.0
	 */
	public void setBorderColor(Color borderColor) {
		setAttr(DAttrs.borderColor, borderColor);
	}

	/**
	 * Sets the color of the border surrounding each slice.
	 * @since 7.2.1.0
	 */
	public void setBorderColor(String borderColor) {
		setBorderColor(new Color(borderColor));
	}

	/**
	 * Sets the color of the border surrounding each slice.
	 * @since 7.2.1.0
	 */
	public void setBorderColor(LinearGradient borderColor) {
		setBorderColor(new Color(borderColor));
	}

	/**
	 * Sets the color of the border surrounding each slice.
	 * @since 7.2.1.0
	 */
	public void setBorderColor(RadialGradient borderColor) {
		setBorderColor(new Color(borderColor));
	}

	/**
	 * Returns the border radius of the series.
	 * @since 7.2.1.0
	 */
	public Number getBorderRadius() {
		return getAttr(DAttrs.borderRadius, null).asNumber();
	}

	/**
	 * Sets the border radius of the series.
	 * @since 7.2.1.0
	 */
	public void setBorderRadius(Number borderRadius) {
		setAttr(DAttrs.borderRadius, borderRadius);
	}

	/**
	 * Returns the border width of the series.
	 */
	public Number getBorderWidth() {
		return getAttr(DAttrs.borderWidth, null).asNumber();
	}

	/**
	 * Sets the border width of the series.
	 */
	public void setBorderWidth(Number borderWidth) {
		setAttr(DAttrs.borderWidth, borderWidth);
	}

	/**
	 * Returns the center of the pie chart relative to the plot area.
	 * Can be percentages or pixel values.
	 * The default behaviour is to center the pie so that all slices and data labels are within the plot area.
	 * As a consequence, the pie may actually jump around in a chart with dynamic values,
	 * as the data labels move. In that case, the center should be explicitly set,
	 * for example to ["50%", "50%"].
	 * <p>
	 * Default: [null, null]
	 * @since 7.2.1.0
	 */
	public List<String> getCenter() {
		if (!containsKey(DAttrs.center)) {
			setCenter(Arrays.asList(null, null));
		}
		return Generics.cast(getAttr(DAttrs.center));
	}

	/**
	 * Sets the center of the pie chart relative to the plot area.
	 * Can be percentages or pixel values.
	 * The default behaviour is to center the pie so that all slices and data labels are within the plot area.
	 * As a consequence, the pie may actually jump around in a chart with dynamic values,
	 * as the data labels move. In that case, the center should be explicitly set,
	 * for example to ["50%", "50%"].
	 * @since 7.2.1.0
	 */
	public void setCenter(List<Object> center) {
		setAttr(DAttrs.center, center, NOT_NULL_VALUE);
	}

	/**
	 * Sets the center of the pie chart relative to the plot area.
	 * Can be percentages or pixel values.
	 * The default behaviour is to center the pie so that all slices and data labels are within the plot area.
	 * As a consequence, the pie may actually jump around in a chart with dynamic values,
	 * as the data labels move. In that case, the center should be explicitly set,
	 * for example to ["50%", "50%"].
	 * @since 7.2.1.0
	 */
	public void setCenter(Object... center) {
		setCenter(Arrays.asList(center));
	}

	/**
	 * Sets the plot options to the series.
	 */
	public void setPlotOptions(SeriesPlotOptions plotOptions) {
		SeriesPlotOptions old = this.plotOptions;
		if (!Objects.equals(old, plotOptions)) {
			if (old != null)
				old.removeOptionDataListener(this);

			if (plotOptions != null) {
				plotOptions.addOptionDataListener(this);
				this.plotOptions = plotOptions;
			}
		}
	}

	/**
	 * Returns the plot options to the series. If null, it will create a new
	 * one.
	 */
	public SeriesPlotOptions getPlotOptions() {
		if (plotOptions == null) {
			plotOptions = new SeriesPlotOptions();
			setPlotOptions(plotOptions);
		}
		return plotOptions;
	}

	/**
	 * Add a point at the end of the current points list.
	 * 
	 * @param point
	 */
	public void addPoint(Point point) {
		addPoint(point, true, false, true);
	}

	/**
	 * Add a point to the end of the current points list.
	 * 
	 * @param point
	 *            the point
	 * @param redraw
	 *            whether to redraw the chart
	 * @param shift
	 *            If shift is true, a point is shifted off the start of the
	 *            series as one is appended to the end.
	 * @param animation
	 *            Whether to apply animation, and optionally animation
	 *            configuration
	 */
	public void addPoint(Point point, boolean redraw, boolean shift,
			boolean animation) {
		addPoint(point, redraw, shift, animation ? new Animation()
				: Animation.NONE);
	}

	/**
	 * Add a point at the end of the current points list.
	 * 
	 * @param point
	 *            the point
	 * @param redraw
	 *            whether to redraw the chart
	 * @param shift
	 *            If shift is true, a point is shifted off the start of the
	 *            series as one is appended to the end.
	 * @param animation
	 *            Whether to apply animation, and optionally animation
	 *            configuration
	 */
	public void addPoint(final Point point, final boolean redraw,
			final boolean shift, final Animation animation) {
		List<Point> list = getData();
		if (list == null)
			list = new LinkedList<Point>();
		list.add(point);

		// monitor if data changed.
		point.addOptionDataListener(this);
		point.setSeries(this);
		setAttr(Attrs.data, list);

		fireEvent(EventType.ADDED, Attrs.point.toString(), point,
				new DeferredCall() {
					public void execute(JSFunction func) {
						func.callFunction("addPoint", point, redraw, shift,
								animation);
					}
				});
	}

	/**
	 * Add a number point to the end of the current points list.
	 */
	public void addPoint(Number y) {
		if (y == null)
			addPoint(new Point());
		else
			addPoint(new Point(y));
	}

	/**
	 * Add a number point to the end of the current points list.
	 */
	public void addPoint(double y) {
		addPoint(new Point(y));
	}

	/**
	 * Add a point with x and y numbers to the end of the current points list.
	 */
	public void addPoint(Number x, Number y) {
		addPoint(new Point(x, y));
	}

	/**
	 * Add a point with x and y numbers to the end of the current points list.
	 */
	public void addPoint(double x, double y) {
		addPoint(new Point(x, y));
	}

	/**
	 * Add a point with x, low, and high numbers to the end of the current
	 * points list.
	 */
	public void addPoint(Number x, Number low, Number high) {
		addPoint(new Point(x, low, high));
	}

	/**
	 * Add a point with x, low, and high numbers to the end of the current
	 * points list.
	 */
	public void addPoint(double x, double low, double high) {
		addPoint(new Point(x, low, high));
	}

	/**
	 * Add a point with name and number to the end of the current points list.
	 */
	public void addPoint(String name, Number y) {
		addPoint(new Point(name, y));
	}

	/**
	 * Add a point with name, Y number, and X index to the end of the current points list.
	 * @since 1.1
	 */
	public void addPoint(String name, Number y, Number x) {
		addPoint(new Point(name, y, x));
	}

	/**
	 * Add a point with name and number to the end of the current points list.
	 */
	public void addPoint(String name, double y) {
		addPoint(new Point(name, y));
	}

	/**
	 * Add a point with from and to to the end of the current points list.
	 * @since 7.2.1.0
	 */
	public void addPoint(String from, String to) {
		addPoint(new Point(from, to));
	}

	/**
	 * Add a point with from, to and weight to the end of the current points list.
	 * @since 7.2.1.0
	 */
	public void addPoint(String from, String to, Number weight) {
		addPoint(new Point(from, to, weight));
	}

	/**
	 * Sets a list of point data
	 */
	public void setData(List<? extends Number> data) {
		Point[] points = new Point[data.size()];
		int i = 0;
		for (Number t : data) {
			if (t == null)
				points[i++] = new Point();
			else
				points[i++] = new Point(t);
		}
		setData(points);
	}

	/**
	 * Sets an array of point data
	 */
	public void setData(Point... data) {
		for (Point p : data) {
			// monitor if data changed.
			p.addOptionDataListener(this);
			p.setSeries(this);
		}

		// use ArrayList instead of fixed ArrayList
		final List<Point> list = new ArrayList<Point>(Arrays.asList(data));
		setAttr(Attrs.data, list);

		fireEvent(EventType.CHANGED, Attrs.data.toString(), data,
				new DeferredCall() {
					public void execute(JSFunction func) {
						func.callFunction("setData", list);
					}
				});
	}

	/**
	 * Sets an array of number data
	 */
	public void setData(Number... data) {
		setData(Arrays.asList(data));
	}

	/**
	 * Sets an array of number data
	 */
	public void setData(Double... data) {
		setData(Arrays.asList(data));
	}

	/**
	 * Sets an array of number data
	 */
	public void setData(Integer... data) {
		setData(Arrays.asList(data));
	}

	/**
	 * Returns the list of point data
	 */
	public List<Point> getData() {
		return Generics.cast(getAttr(Attrs.data, null).asValue());
	}

	/**
	 * Returns the point in the list from the given index.
	 */
	public Point getPoint(int index) {
		return getData().get(index);
	}

	/**
	 * Sets the index of the series in the chart, affecting the internal index
	 * in the chart.series array, the visible Z index as well as the order in
	 * the legend.
	 */
	public void setIndex(int index) {
		if (index < 0) {
			removeKey(DAttrs.index, true);
		} else {
			setAttr(DAttrs.index, index);
		}
	}

	/**
	 * Returns the index of the series in the chart, affecting the internal
	 * index in the chart.series array, the visible Z index as well as the order
	 * in the legend.
	 * <p>
	 * Default: -1, it means depended on the implementaion.
	 */
	public int getIndex() {
		return getAttr(DAttrs.index, -1).asInt();
	}

	/**
	 * Returns the length of the bell curve.
	 * A unit of the length of the bell curve is standard deviation.
	 *
	 * Requires bellcurve.
	 * <p>
	 * Default: 3
	 * @since 7.2.1.0
	 */
	public Number getIntervals() {
		return getAttr(DAttrs.intervals, 3).asNumber();
	}

	/**
	 * Sets the length of the bell curve.
	 * A unit of the length of the bell curve is standard deviation.
	 *
	 * Requires bellcurve.
	 * @since 7.2.1.0
	 */
	public void setIntervals(Number intervals) {
		setAttr(DAttrs.intervals, intervals, 3);
	}

	/**
	 * In circular view, returns the size of the inner diameter of the circle.
	 * Can be a percentage or pixel value.
	 * Percentages are relative to the outer perimeter.
	 * Pixel values are given as integers.
	 *
	 * If the {@link #getRows()} option is set, it overrides the innerSize setting.
	 * <p>
	 * Default: "40%"
	 * @since 7.2.1.0
	 */
	public Object getInnerSize() {
		return getAttr(DAttrs.innerSize, "40%").asValue();
	}

	/**
	 * In circular view, sets the size of the inner diameter of the circle.
	 * Can be a percentage.
	 * Percentages are relative to the outer perimeter.
	 *
	 * If the {@link #getRows()} option is set, it overrides the innerSize setting.
	 * @since 7.2.1.0
	 */
	public void setInnerSize(String innerSize) {
		setAttr(DAttrs.innerSize, innerSize);
	}

	/**
	 * In circular view, sets the size of the inner diameter of the circle.
	 * Can be a pixel value.
	 * Pixel values are given as integers.
	 *
	 * If the {@link #getRows()} option is set, it overrides the innerSize setting.
	 */
	public void setInnerSize(int innerSize) {
		setAttr(DAttrs.innerSize, innerSize);
	}

	/**
	 * Sets the option of data labels.
	 * 
	 * @param dataLabels
	 * @see SeriesPlotOptions#setDataLabels(DataLabels)
	 */
	public void setDataLabels(DataLabels dataLabels) {
		getPlotOptions().setDataLabels(dataLabels);
	}

	/**
	 * Returns the option of data labels.
	 * 
	 * @see SeriesPlotOptions#getDataLabels()
	 */
	public DataLabels getDataLabels() {
		return getPlotOptions().getDataLabels();
	}

	/**
	 * Returns individual color for the series. By default the color is pulled
	 * from the global <code>colors</code> array.
	 * <p>
	 * Default: null.
	 */
	public Color getColor() {
		return (Color) getAttr(DAttrs.color, null).asValue();
	}

	/**
	 * Sets individual color for the series. By default the color is pulled from
	 * the global <code>colors</code> array.
	 */
	public void setColor(Color color) {
		setAttr(DAttrs.color, color);
	}

	/**
	 * Sets individual color for the series. By default the color is pulled from
	 * the global <code>colors</code> array.
	 */
	public void setColor(String color) {
		setColor(new Color(color));
	}

	/**
	 * Sets individual color for the series. By default the color is pulled from
	 * the global <code>colors</code> array.
	 */
	public void setColor(LinearGradient color) {
		setColor(new Color(color));
	}

	/**
	 * Sets individual color for the series. By default the color is pulled from
	 * the global <code>colors</code> array.
	 */
	public void setColor(RadialGradient color) {
		setColor(new Color(color));
	}

	/**
	 * When using dual or multiple color axes, this defines which colorAxis
	 * the particular series is connected to. It refers to either the {@link ColorAxis#getId()}
	 * or the index of the axis in the colorAxis array, with 0 being the first.
	 * Set this option to false to prevent a series from connecting to the default color axis.
	 * <p>
	 * Default: 0
	 * @since 7.2.1.0
	 */
	public Object getColorAxis() {
		return getAttr(DAttrs.colorAxis, 0).asValue();
	}

	/**
	 * When using dual or multiple color axes, this defines which colorAxis
	 * the particular series is connected to.
	 * It refers to the index of the axis in the colorAxis array, with 0 being the first.
	 * @since 7.2.1.0
	 */
	public void setColorAxis(Number colorAxis) {
		setAttr(DAttrs.colorAxis, colorAxis, 0);
	}

	/**
	 * When using dual or multiple color axes, this defines which colorAxis
	 * the particular series is connected to. It refers to the {@link ColorAxis#getId().
	 * @since 7.2.1.0
	 */
	public void setColorAxis(String colorAxis) {
		setAttr(DAttrs.colorAxis, colorAxis);
	}

	/**
	 * When using dual or multiple color axes, this defines which colorAxis
	 * the particular series is connected to.
	 * Set this option to false to prevent a series from connecting to the default color axis.
	 * @since 7.2.1.0
	 */
	public void setColorAxis(boolean colorAxis) {
		setAttr(DAttrs.colorAxis, colorAxis);
	}

	/**
	 * Returns what data value should be used to calculate point color if colorAxis is used.
	 * Requires to set min and max if some custom point property is used or if
	 * approximation for data grouping is set to 'sum'.
	 * <p>
	 * Default: "y"
	 * @since 7.2.1.0
	 */
	public String getColorKey() {
		return getAttr(DAttrs.colorKey, "y").asString();
	}

	/**
	 * Sets what data value should be used to calculate point color if colorAxis is used.
	 * Requires to set min and max if some custom point property is used or if
	 * approximation for data grouping is set to 'sum'.
	 * @since 7.2.1.0
	 */
	public void setColorKey(String colorKey) {
		setAttr(DAttrs.colorKey, colorKey, "y");
	}

	/**
	 * Returns whether color by point option is enabled or not
	 */
	public boolean isColorByPoint() {
		return getAttr(DAttrs.colorByPoint, false).asBoolean();
	}
	
	/**
	 * Sets whether to enable the color by point option or not
	 */
	public void setColorByPoint(boolean colorByPoint) {
		setAttr(DAttrs.colorByPoint, colorByPoint);
	}

	/**
	 * Returns a series specific or series type specific color set to apply
	 * instead of the global {@link Charts#getColors()} when
	 * {@link #setColorByPoint(boolean)} is true.
	 * @since 7.2.1.0
	 */
	public List<Color> getColors() {
		return Generics.cast(getAttr(DAttrs.colors, null).asValue());
	}

	/**
	 * Sets a series specific or series type specific color set to apply
	 * instead of the global {@link Charts#getColors()} when
	 * {@link #setColorByPoint(boolean)} is true.
	 * @since 7.2.1.0
	 */
	public void setColors(List<Color> colors) {
		setAttr(DAttrs.colors, colors);
	}

	/**
	 * Sets a series specific or series type specific color set to apply
	 * instead of the global {@link Charts#getColors()} when
	 * {@link #setColorByPoint(boolean)} is true.
	 * @since 7.2.1.0
	 */
	public void setColors(String... source) {
		Color[] colors = new Color[source.length];
		int i = 0;
		for (String s : source)
			colors[i++] = new Color(s);
		setColors(Arrays.asList(colors));
	}

	/**
	 * Returns you can set the cursor to "pointer" if you have click events
	 * attached to the series, to signal to the user that the points and lines
	 * can be clicked.
	 * <p>
	 * Default: null
	 * @since 7.2.1.0
	 */
	public String getCursor() {
		return getAttr(DAttrs.cursor, null).asString();
	}

	/**
	 * Sets you can set the cursor to "pointer" if you have click events
	 * attached to the series, to signal to the user that the points and lines
	 * can be clicked.
	 * @since 7.2.1.0
	 */
	public void setCursor(String cursor) {
		setAttr(DAttrs.cursor, cursor);
	}

	/**
	 * Returns the options of the draggable-points module allows points
	 * to be moved around or modified in the chart, if any. Otherwise, create a new one.
	 * In addition to the options mentioned under the dragDrop API structure,
	 * the module fires three events, point.dragStart, point.drag and point.drop.
	 * @since 7.2.1.0
	 */
	public DragDrop getDragDrop() {
		DragDrop dragDrop = (DragDrop) getAttr(DAttrs.dragDrop);
		if (dragDrop == null) {
			dragDrop = new DragDrop();
			setDragDrop(dragDrop);
		}
		return dragDrop;
	}

	/**
	 * Sets the options of the draggable-points module allows points
	 * to be moved around or modified in the chart.
	 * In addition to the options mentioned under the dragDrop API structure,
	 * the module fires three events, point.dragStart, point.drag and point.drop.
	 * @since 7.2.1.0
	 */
	public void setDragDrop(DragDrop dragDrop) {
		setAttr(DAttrs.dragDrop, dragDrop);
	}

	/**
	 * Returns enable or disable the mouse tracking for a specific series. This
	 * includes point tooltips and click events on graphs and points. For large
	 * datasets it improves performance.
	 * @since 7.2.1.0
	 */
	public boolean isEnableMouseTracking() {
		return getAttr(DAttrs.enableMouseTracking, true).asBoolean();
	}

	/**
	 * Sets enable or disable the mouse tracking for a specific series. This
	 * includes point tooltips and click events on graphs and points. For large
	 * datasets it improves performance.
	 * @since 7.2.1.0
	 */
	public void setEnableMouseTracking(boolean enableMouseTracking) {
		setAttr(DAttrs.enableMouseTracking, enableMouseTracking);
	}

	/**
	 * In circular view, returns the end angle of the item layout, in degrees where 0 is up.
	 * Default: null
	 * @since 7.2.1.0
	 */
	public Number getEndAngle() {
		return getAttr(DAttrs.endAngle, null).asNumber();
	}

	/**
	 * In circular view, sets the end angle of the item layout, in degrees where 0 is up.
	 * @since 7.2.1.0
	 */
	public void setEndAngle(Number endAngle) {
		setAttr(DAttrs.endAngle, endAngle);
	}

	/**
	 * Returns fill color or gradient for the area.
	 * When null, the series' color is used with the series' fillOpacity.
	 * <p>
	 * Default: null
	 * @since 7.2.1.0
	 */
	public Color getFillColor() {
		return (Color) getAttr(DAttrs.fillColor, null).asValue();
	}

	/**
	 * Sets fill color for the area.
	 * When null, the series' color is used with the series' fillOpacity.
	 * @since 7.2.1.0
	 */
	public void setFillColor(Color color) {
		setAttr(DAttrs.fillColor, color, NOT_NULL_VALUE);
	}

	/**
	 * Sets fill color for the area.
	 * When null, the series' color is used with the series' fillOpacity.
	 * @since 7.2.1.0
	 */
	public void setFillColor(String color) {
		setFillColor(new Color(color));
	}

	/**
	 * Sets fill gradient for the area.
	 * When null, the series' color is used with the series' fillOpacity.
	 * @since 7.2.1.0
	 */
	public void setFillColor(LinearGradient color) {
		setFillColor(new Color(color));
	}

	/**
	 * Sets fill gradient for the area.
	 * When null, the series' color is used with the series' fillOpacity.
	 * @since 7.2.1.0
	 */
	public void setFillColor(RadialGradient color) {
		setFillColor(new Color(color));
	}

	public int getGapSize() {
		return getAttr(DAttrs.gapSize, 0).asInt();
	}

	public void setGapSize(int gapSize) {
		setAttr(DAttrs.gapSize, gapSize, 0);
	}

	/**
	 * Returns the layout of the items in rectangular view.
	 * Can be either horizontal or vertical
	 * <p>
	 * Default: "vertical"
	 * @since 7.2.1.0
	 */
	public String getLayout() {
		return getAttr(DAttrs.layout, "vertical").asString();
	}

	/**
	 * Sets the layout of the items in rectangular view.
	 * Can be either horizontal or vertical
	 * @since 7.2.1.0
	 */
	public void setLayout(String layout) {
		if (!("horizontal".equals(layout) || "vertical".equals(layout)))
			throw new IllegalArgumentException("Unsupported layout: [ " + layout + " ]");
		setAttr(DAttrs.layout, layout);
	}

	/**
	 * Sets the sequential index of the series in the legend.
	 */
	public void setLegendIndex(int legendIndex) {
		setAttr(DAttrs.legendIndex, legendIndex);
	}

	/**
	 * Returns the sequential index of the series in the legend.
	 */
	public int getLegendIndex() {
		return getAttr(DAttrs.legendIndex, -1).asInt();
	}
	
	/**
	 * Returns the width of the point marker's outline.
	 */
	public Number getLineWidth() {
		return getAttr(DAttrs.lineWidth, null).asNumber();
	}

	/**
	 * Sets the width of the point marker's outline.
	 */
	public void setLineWidth(Number lineWidth) {
		setAttr(DAttrs.lineWidth, lineWidth, null);
	}

	/**
	 * Returns the link style options, if any. Otherwise, create a new one.
	 * @since 7.2.1.0
	 */
	public NetworkGraphLink getLink() {
		NetworkGraphLink link = (NetworkGraphLink) getAttr(DAttrs.link);
		if (link == null) {
			link = new NetworkGraphLink();
			setLink(link);
		}
		return link;
	}

	/**
	 * Sets the link style options.
	 * @since 7.2.1.0
	 */
	public void setLink(NetworkGraphLink link) {
		setAttr(DAttrs.link, link);
	}

	/**
	 * Returns the marker for this series.
	 */
	public Marker getMarker() {
		Marker marker = (Marker) getAttr(DAttrs.marker);
		if (marker == null) {
			marker = new Marker();
			setAttr(DAttrs.marker, marker);
		}
		return marker;
	}

	/**
	 * Sets the marker for this series.
	 */
	public void setMarker(Marker marker) {
		setAttr(DAttrs.marker, marker);
	}

	/**
	 * Returns maximum bubble size.
	 * Bubbles will automatically size between the minSize and maxSize to reflect the z value of each bubble.
	 * Can be either pixels (when no unit is given), or a percentage of the smallest one of the plot width and height.
	 * <p>
	 * Default: "20%"
	 * @since 7.2.1.0
	 */
	public Object getMaxSize() {
		return getAttr(DAttrs.maxSize, "20%");
	}

	/**
	 * Sets maximum bubble size.
	 * Bubbles will automatically size between the minSize and maxSize to reflect the z value of each bubble.
	 * Can be pixels (when no unit is given).
	 * @since 7.2.1.0
	 */
	public void setMaxSize(Number maxSize) {
		setAttr(DAttrs.maxSize, maxSize);
	}

	/**
	 * Sets maximum bubble size.
	 * Bubbles will automatically size between the minSize and maxSize to reflect the z value of each bubble.
	 * Can be a percentage of the smallest one of the plot width and height.
	 * @since 7.2.1.0
	 */
	public void setMaxSize(String maxSize) {
		setAttr(DAttrs.maxSize, maxSize);
	}

	/**
	 * Returns the minimum size.
	 * @since 7.2.1.0
	 */
	public Object getMinSize() {
		return getAttr(DAttrs.minSize);
	}

	/**
	 * Sets the minimum size.
	 * @since 7.2.1.0
	 */
	public void setMinSize(Number minSize) {
		setAttr(DAttrs.minSize, minSize);
	}

	/**
	 * Sets the minimum size.
	 * @since 7.2.1.0
	 */
	public void setMinSize(String minSize) {
		setAttr(DAttrs.minSize, minSize);
	}

	/**
	 * Sets the name of the series as shown in the legend, tooltip etc.
	 */
	public void setName(String name) {
		setAttr(DAttrs.name, name);
	}

	/**
	 * Returns the id of another series in the chart that the wind barbs are projected on.
	 * When null, the wind symbols are drawn on the X axis, but offset up or down by the yOffset setting.
	 * <p>
	 * Default: null
	 * @since 7.2.1.0
	 */
	public String getOnSeries() {
		return getAttr(DAttrs.onSeries, null).asString();
	}

	/**
	 * Sets the id of another series in the chart that the wind barbs are projected on.
	 * When null, the wind symbols are drawn on the X axis, but offset up or down by the yOffset setting.
	 * @since 7.2.1.0
	 */
	public void setOnSeries(String onSeries) {
		setAttr(DAttrs.onSeries, onSeries);
	}
	
	/**
	 * Returns the amount of degrees that are allowed to overshoot
	 * Default: null
	 */
	public Number getOvershoot() {
		return getAttr(DAttrs.overshoot, null).asNumber();
	}
	
	/**
	 * Sets the amount of degrees that are allowed to overshoot
	 */
	public void setOvershoot(Number overshoot) {
		setAttr(DAttrs.overshoot, overshoot);
	}

	/**
	 * Returns the point interval of the series.
	 * Default: null
	 */
	public Number getPointInterval() {
		return getAttr(DAttrs.pointInterval, null).asNumber();
	}
	
	/**
	 * Sets the point interval of the series.
	 */
	public void setPointInterval(Number pointInterval) {
		setAttr(DAttrs.pointInterval, pointInterval);
	}
	
	/**
	 * On datetime series, this allows for setting the pointInterval to the two
	 * irregular time units, month and year. Combine it with pointInterval to
	 * draw quarters, 6 months, 10 years etc.
	 * <p>
	 * Default: null
	 * @since 2.1.0
	 */
	public String getPointIntervalUnit() {
		return getAttr(DAttrs.pointIntervalUnit, null).asString();
	}
	/**
	 * Set the irregular time unit for point interval, only "month" or "year" is allowed.
	 * @param pointIntervalUnit either "month" or "year"
	 * @since 2.1.0
	 */
	public void setPointIntervalUnit(String pointIntervalUnit) {
		if (!("month".equalsIgnoreCase(pointIntervalUnit) || "year".equalsIgnoreCase(pointIntervalUnit))) {
			throw new IllegalArgumentException("Unsupported point interval unit:[" + pointIntervalUnit + "]");
		}
		setAttr(DAttrs.pointIntervalUnit, pointIntervalUnit);
	}
	
	/**
	 * Returns the padding between each column or bar, in x axis units.
	 * Default: 0.1
	 */
	public Number getPointPadding() {
		return getAttr(DAttrs.pointPadding, 0.1).asNumber();
	}
	
	/**
	 * Sets the padding between each column or bar, in x axis units.
	 */
	public void setPointPadding(Number pointPadding) {
		setAttr(DAttrs.pointPadding, pointPadding, 0.1);
	}
	
	/**
	 * Returns the point placement.
	 * <p>
	 * Possible values: null, "on", "between" or numeric.
	 * </p>
	 * <p>
	 * In a column chart, when pointPlacement is "on", the point will not create
	 * any padding of the X axis. In a polar column chart this means that the
	 * first column points directly north. If the pointPlacement is "between",
	 * the columns will be laid out between ticks. This is useful for example
	 * for visualising an amount between two points in time or in a certain
	 * sector of a polar chart.
	 * </p>
	 * <p>
	 * The point placement can also be numeric, where 0 is on the axis value,
	 * -0.5 is between this value and the previous, and 0.5 is between this
	 * value and the next. Unlike the textual options, numeric point placement
	 * options won't affect axis padding.
	 * </p>
	 * <p>
	 * Defaults to <code>null</code> in cartesian charts, <code>"between"</code>
	 * in polar charts.
	 */
	public Object getPointPlacement() {
		return getAttr(DAttrs.pointPlacement, null).asValue();
	}

	/**
	 * Sets the point placement.
	 * <p>
	 * Possible values: null, "on", "between".
	 * </p>
	 * <p>
	 * In a column chart, when pointPlacement is "on", the point will not create
	 * any padding of the X axis. In a polar column chart this means that the
	 * first column points directly north. If the pointPlacement is "between",
	 * the columns will be laid out between ticks. This is useful for example
	 * for visualising an amount between two points in time or in a certain
	 * sector of a polar chart.
	 * <p>
	 * Defaults to <code>null</code> in cartesian charts, <code>"between"</code>
	 * in polar charts.
	 */
	public void setPointPlacement(String pointPlacement) {
		if (!("on".equals(pointPlacement) || "between".equals(pointPlacement)))
			throw new IllegalArgumentException("Unsupported pointPlacement: ["
					+ pointPlacement + "]");
		setAttr(DAttrs.pointPlacement, pointPlacement);
	}

	/**
	 * Sets the point placement with a number
	 * <p>
	 * The point placement can also be numeric, where 0 is on the axis value,
	 * -0.5 is between this value and the previous, and 0.5 is between this
	 * value and the next. Unlike the textual options, numeric point placement
	 * options won't affect axis padding.
	 * </p>
	 */
	public void setPointPlacement(Number pointPlacement) {
		setAttr(DAttrs.pointPlacement, pointPlacement);
	}

	/**
	 * Returns how many points should be plotted within 1 interval.
	 * See plotOptions.bellcurve.intervals.
	 *
	 * Requires bellcurve.
	 * <p>
	 * Default: 3
	 * @since 7.2.1.0
	 */
	public Number getPointsInInterval() {
		return getAttr(DAttrs.pointsInInterval, 3).asNumber();
	}

	/**
	 * Sets how many points should be plotted within 1 interval.
	 * See plotOptions.bellcurve.intervals.
	 *
	 * Requires bellcurve.
	 * @since 7.2.1.0
	 */
	public void setPointsInInterval(Number pointsInInterval) {
		setAttr(DAttrs.pointsInInterval, pointsInInterval, 3);
	}

	/**
	 * If no x values are given for the points in a series, pointStart defines
	 * on what value to start. For example, if a series contains one yearly
	 * value starting from 1945, set pointStart to 1945.
	 *
	 * Requires bellcurve.
	 * <p>
	 * Default: 0
	 * @since 7.2.1.0
	 */
	public Number getPointStart() {
		return getAttr(DAttrs.pointStart, 0).asNumber();
	}

	/**
	 * If no x values are given for the points in a series, pointStart defines
	 * on what value to start. For example, if a series contains one yearly
	 * value starting from 1945, set pointStart to 1945.
	 *
	 * Requires bellcurve.
	 * @since 7.2.1.0
	 */
	public void setPointStart(Number pointStart) {
		setAttr(DAttrs.pointStart, pointStart, 0);
	}

	/**
	 * Default: null
	 * @since 7.2.1.0
	 */
	public Number getPointRange() {
		return getAttr(DAttrs.pointRange, null).asNumber();
	}

	/**
	 * @since 7.2.1.0
	 */
	public void setPointRange(Number pointRange) {
		setAttr(DAttrs.pointRange, pointRange);
	}

	/**
	 * Returns the name of the series as shown in the legend, tooltip etc.
	 */
	public String getName() {
		return getAttr(DAttrs.name, null).asString();
	}

	/**
	 * Returns the size of nodes.
	 * @since 7.2.1.0
	 */
	public int getNodesSize() {
		List<Series> list = Generics.cast(getAttr(DAttrs.nodes));
		return list == null ? 0 : list.size();
	}

	/**
	 * Returns the Node options, if any. Otherwise, create a new one.
	 * <p>
	 * Returns the first node (index 0) by default.
	 *
	 * @see Node
	 * @see #getNodes(int)
	 * @since 7.2.1.0
	 */
	public Node getNodes() {
		return getNodes(0);
	}

	/**
	 * Returns the Node options from the given index, if any. Otherwise,
	 * create a new one.
	 *
	 * @see Node
	 * @since 7.2.1.0
	 */
	public Node getNodes(int index) {
		List<Node> list = Generics.cast(getAttr(DAttrs.nodes));
		if (list == null) {
			list = new LinkedList<Node>();
			setAttr(DAttrs.nodes, list);
		}

		int size = list.size();
		if (size <= index) {
			for (int i = index - size + 1; i > 0; i--) {
				list.add(new Node());
			}
		}
		return list.get(index);
	}

	/**
	 * Returns the number of rows to display in the rectangular or circular view.
	 * If the {@link #getInnerSize()} is set, it will be overridden by the rows setting.
	 * <p>
	 * Default: null
	 * @since 7.2.1.0
	 */
	public Number getRows() {
		return getAttr(DAttrs.rows, null).asNumber();
	}

	/**
	 * Sets the number of rows to display in the rectangular or circular view.
	 * If the {@link #getInnerSize()} is set, it will be overridden by the rows setting.
	 * @since 7.2.1.0
	 */
	public void setRows(Number rows) {
		setAttr(DAttrs.rows, rows);
	}

	/**
	 * Returns whether to display this particular series or series type in the legend.
	 * Standalone series are shown in legend by default, and linked series are not.
	 * It is possible to show series that use colorAxis by setting this option to true.
	 * <p>
	 * Default: true
	 * @since 7.2.1.0
	 */
	public boolean isShowInLegend() {
		return getAttr(DAttrs.showInLegend, true).asBoolean();
	}

	/**
	 * Sets whether to display this particular series or series type in the legend.
	 * Standalone series are shown in legend by default, and linked series are not.
	 * It is possible to show series that use colorAxis by setting this option to true.
	 * @since 7.2.1.0
	 */
	public void setShowInLegend(boolean showInLegend) {
		setAttr(DAttrs.showInLegend, showInLegend);
	}

	/**
	 * Returns the diameter of the pie relative to the plot area.
	 * Can be a percentage or pixel value. Pixel values are given as integers.
	 * The default behaviour is to scale to the plot area and give room for data labels
	 * within the plot area. slicedOffset is also included in the default size calculation.
	 * As a consequence, the size of the pie may vary when points are updated and data labels more around.
	 * In that case it is best to set a fixed value, for example "75%".
	 * <p>
	 * Default: null
	 * @since 7.2.1.0
	 */
	public Object getSize() {
		return getAttr(DAttrs.size);
	}

	/**
	 * Sets the diameter of the pie relative to the plot area.
	 * Can be a percentage or pixel value. Pixel values are given as integers.
	 * The default behaviour is to scale to the plot area and give room for data labels
	 * within the plot area. slicedOffset is also included in the default size calculation.
	 * As a consequence, the size of the pie may vary when points are updated and data labels more around.
	 * In that case it is best to set a fixed value, for example "75%".
	 * @since 7.2.1.0
	 */
	public void setSize(Number size) {
		setAttr(DAttrs.size, size);
	}

	/**
	 * Sets the diameter of the pie relative to the plot area.
	 * Can be a percentage or pixel value. Pixel values are given as integers.
	 * The default behaviour is to scale to the plot area and give room for data labels
	 * within the plot area. slicedOffset is also included in the default size calculation.
	 * As a consequence, the size of the pie may vary when points are updated and data labels more around.
	 * In that case it is best to set a fixed value, for example "75%".
	 * @since 7.2.1.0
	 */
	public void setSize(String size) {
		setAttr(DAttrs.size, size);
	}

	/**
	 * Returns what the plot area value should be represented by.
	 * The default, area, corresponds best to the human perception of the size of each area.
	 * <p>
	 * Default: "area"
	 * @since 7.2.1.0
	 */
	public String getSizeBy() {
		return getAttr(DAttrs.sizeBy, "area").asString();
	}

	/**
	 * Sets what the plot area value should be represented by.
	 * @since 7.2.1.0
	 */
	public void setSizeBy(String sizeBy) {
		setAttr(DAttrs.sizeBy, sizeBy);
	}

	/**
	 * Sets the option allows grouping series in a stacked chart. The stack
	 * option can be a string or a number or anything else, as long as the
	 * grouped series' stack options match each other.
	 */
	public void setStack(String stack) {
		setAttr(DAttrs.stack, stack);
	}
	
	/**
	 * Sets the option allows grouping series in a stacked chart. The stack
	 * option can be a string or a number or anything else, as long as the
	 * grouped series' stack options match each other.
	 */
	public void setStack(Number stack) {
		setAttr(DAttrs.stack, stack);
	}

	/**
	 * Returns the option allows grouping series in a stacked chart. The stack
	 * option can be a string or a number or anything else, as long as the
	 * grouped series' stack options match each other.
	 */
	public Object getStack() {
		return getAttr(DAttrs.stack, null).asValue();
	}

	/**
	 * In circular view, returns the start angle of the item layout, in degrees where 0 is up.
	 * Default: null
	 * @since 7.2.1.0
	 */
	public Number getStartAngle() {
		return getAttr(DAttrs.startAngle, null).asNumber();
	}

	/**
	 * In circular view, sets the start angle of the item layout, in degrees where 0 is up.
	 * @since 7.2.1.0
	 */
	public void setStartAngle(Number startAngle) {
		setAttr(DAttrs.startAngle, startAngle);
	}

	/**
	 * Returns all options related with look and position of targets, if any.
	 * Otherwise, create a new one.
	 * @since 7.2.1.0
	 */
	public TargetOptions getTargetOptions() {
		TargetOptions targetOptions = (TargetOptions) getAttr(DAttrs.targetOptions);
		if (targetOptions == null) {
			targetOptions = new TargetOptions();
			setTargetOptions(targetOptions);
		}
		return targetOptions;
	}

	/**
	 * Sets all options related with look and position of targets.
	 * @since 7.2.1.0
	 */
	public void setTargetOptions(TargetOptions targetOptions) {
		setAttr(DAttrs.targetOptions, targetOptions);
	}

	/**
	 * Returns the tooltip options.
	 */
	public Tooltip getTooltip() {
		Tooltip tooltip = (Tooltip) getAttr(DAttrs.tooltip);
		if (tooltip == null) {
			tooltip = new Tooltip();
			setTooltip(tooltip);
		}
		return tooltip;
	}
	
	/**
	 * Sets the tooltip options.
	 */
	public void setTooltip(Tooltip tooltip) {
		setAttr(DAttrs.tooltip, tooltip);
	}
	
	
	/**
	 * The type of series. Can be one of area, areaspline, bar, column, line,
	 * pie, scatter, spline, arearange, areasplinerange and columnrange.
	 */
	public void setType(String type) {
		setAttr(DAttrs.type, type);
	}

	/**
	 * The type of series. Can be one of area, areaspline, bar, column, line,
	 * pie, scatter, spline, arearange, areasplinerange and columnrange.
	 */
	public String getType() {
		return getAttr(DAttrs.type, null).asString();
	}

	/**
	 * When using dual or multiple x axes, this number defines which xAxis the
	 * particular series is connected to. It refers to either the axis id or the
	 * index of the axis in the xAxis array, with 0 being the first. Defaults to
	 * 0.
	 */
	public void setXAxis(int xAxis) {
		setAttr(DAttrs.xAxis, xAxis);
	}

	/**
	 * When using dual or multiple x axes, this number defines which xAxis the
	 * particular series is connected to. It refers to either the axis id or the
	 * index of the axis in the xAxis array, with 0 being the first. Defaults to
	 * 0.
	 */
	public void setXAxis(String xAxis) {
		setAttr(DAttrs.xAxis, xAxis);
	}

	/**
	 * When using dual or multiple x axes, this number defines which xAxis the
	 * particular series is connected to. It refers to either the axis id or the
	 * index of the axis in the xAxis array, with 0 being the first. Defaults to
	 * 0.
	 */
	public Object getXAxis() {
		return getAttr(DAttrs.xAxis, 0).asValue();
	}

	/**
	 * When using dual or multiple y axes, this number defines which yAxis the
	 * particular series is connected to. It refers to either the axis id or the
	 * index of the axis in the yAxis array, with 0 being the first. Defaults to
	 * 0.
	 */
	public void setYAxis(int yAxis) {
		setAttr(DAttrs.yAxis, yAxis);
	}

	/**
	 * When using dual or multiple y axes, this number defines which yAxis the
	 * particular series is connected to. It refers to either the axis id or the
	 * index of the axis in the yAxis array, with 0 being the first. Defaults to
	 * 0.
	 */
	public void setYAxis(String yAxis) {
		setAttr(DAttrs.yAxis, yAxis);
	}

	/**
	 * When using dual or multiple y axes, this number defines which yAxis the
	 * particular series is connected to. It refers to either the axis id or the
	 * index of the axis in the yAxis array, with 0 being the first. Defaults to
	 * 0.
	 */
	public Object getYAxis() {
		return getAttr(DAttrs.yAxis, 0).asValue();
	}

	/**
	 * Define the visual z index of the series.
	 */
	public void setZIndex(int zIndex) {
		setAttr(DAttrs.zIndex, zIndex);
	}

	/**
	 * Define the visual z index of the series.
	 */
	public int getZIndex() {
		return getAttr(DAttrs.zIndex, -1).asInt();
	}

	/**
	 * Defines the Axis on which the zones are applied.
	 * @return the zone axis or the default value "y"
	 * @since 2.1.0
	 */
	public String getZoneAxis() {
		return getAttr(DAttrs.zoneAxis, "y").asString();
	}
	/**
	 * Set the axis on which the zones are applied. Only X, Y, or Z axis are supported.
	 * @param zoneAxis either "x", "y", or "z"
	 * @since 2.1.0
	 */
	public void setZoneAxis(String zoneAxis) {
		if (!("x".equalsIgnoreCase(zoneAxis) || "y".equalsIgnoreCase(zoneAxis) || "z".equalsIgnoreCase(zoneAxis))) {
			throw new IllegalArgumentException("Unsupported zone axis: [" + zoneAxis + "]");
		}
		setAttr(DAttrs.zoneAxis, zoneAxis, "y");
	}
	
	/**
	 * An list defining zones within a series. Zones can be applied to the X
	 * axis, Y axis or Z axis for bubbles, according to the zoneAxis option.
	 * @since 2.1.0
	 * @see Zone
	 */
	public List<Zone> getZones() {
		return Generics.cast(getAttr(DAttrs.zones, null).asValue());
	}
	/**
	 * Set the zones within a series.
	 * @since 2.1.0
	 * @see #getZones()
	 * @see Zone
	 */
	public void setZones(List<Zone> zones) {
		setAttr(DAttrs.zones, zones);
	}
	
	/**
	 * Hides the series if visible. If the {@link Chart#isIgnoreHiddenSeries()}
	 * option is true,the chart is redrawn without this series.
	 */
	public void hide() {
		setVisible(false);
	}

	/**
	 * Returns whether the series is visible or not.
	 * <p>
	 * Default: true
	 */
	public boolean isVisible() {
		return getAttr(Attrs.visible, true).asBoolean();
	}

	/**
	 * Sets whether the series is visible.
	 */
	public void setVisible(final boolean visible) {
		if (setVisibleDirectly(visible)) {
			fireEvent(EventType.CHANGED, "visible", this, new DeferredCall() {
				public void execute(JSFunction func) {
					// B-ZKCHARTS-72: set redraw false then call chart.redraw()
					func.callFunction("setVisible", visible, false);
				}
			});
		}
	}

	/**
	 * Sets whether the series is visible directly.
	 */
	public boolean setVisibleDirectly(final boolean visible) {
		return setAttr(Attrs.visible, visible, true);
	}

	/**
	 * Shows the series if hidden.
	 */
	public void show() {
		setVisible(true);
	}

	public void remove() {
		fireEvent(EventType.DESTROYED, "series", this, new DeferredCall() {
			public void execute(JSFunction func) {
				func.callFunction("remove");
			}
		});
		clearOptonDataListener();
	}

	public void select() {
		fireEvent(EventType.CHANGED, "select", this, new DeferredCall() {
			public void execute(JSFunction func) {
				func.callFunction("select");
			}

		});
	}

	public void onChange(OptionDataEvent event) {
		final Optionable target = event.getTarget();
		if (target instanceof Point) {
			final int index = getData().indexOf(target);
			if (event.hasJSFunctionCall()) {
				event.addJSFunctionCall(new DeferredCall() {
					public void execute(JSFunction func) {
						func.callArray("data", index);
					}

				});
			}
			if (event.getType() == OptionDataEvent.EventType.DESTROYED) {
				List<Point> list = getData();
				if (list != null)
					list.remove(index);
				event.setCurrentType(EventType.REMOVED); // ZKCHARTS-56 REMOVED for series
			}
		} else {
			event.addJSFunctionCall(new DeferredCall() {
				public void execute(JSFunction func) {
					if (target instanceof DataLabels) {
						JSONObject json = new JSONObject();
						json.put("dataLabels", target);
						func.callFunction("update", json);
					} else {
						func.callFunction("update", target);
					}
				}
			});
		}
		super.onChange(event);
	}

	public String toJSONString() {
		if (plotOptions != null) {
			// we don't use merge(options) here
			Map<String, AnyVal<Object>> clone = new LinkedHashMap<String, AnyVal<Object>>(
					options);
			clone.putAll(plotOptions.options);
			return JSONObject.toJSONString(clone);
		}
		return super.toJSONString();
	}
}
